# Generated by Django 3.2.20 on 2023-11-06 20:00

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Mock',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('index', models.IntegerField()),
                ('headers', models.JSONField()),
                ('content', models.JSONField()),
                ('status', models.IntegerField()),
                ('method', models.CharField(max_length=6)),
                ('pattern', models.CharField(max_length=1000)),
                ('spec_name', models.CharField(db_index=True, max_length=100)),
            ],
        ),
    ]



import json
import yaml
import os

from django.test import TestCase, override_settings
from django.urls import reverse

from apistubs import settings as su_settings
from apistubs import urls
from apistubs.helpers import load_apistubs_yaml

__all__ = (
    'SpecificationViewTests',
)


PROJECT = 'account'


@override_settings(ROOT_URLCONF=urls, PROJECT=PROJECT)
class SpecificationViewTests(TestCase):
    def get_url(self, spec):
        return reverse('index', kwargs={'spec': spec})

    def get_spec_url(self, spec):
        return reverse('spec', kwargs={'spec': spec})

    def post(self, spec, data, cdn=False):
        data = json.dumps(data)
        if cdn:
            return self.client.post(self.get_spec_url_cdn(spec), data=data, content_type='application/json')
        return self.client.post(self.get_url(spec), data=data, content_type='application/json')

    def get(self, spec, cdn=False):
        if cdn:
            return self.client.get(self.get_spec_url_cdn(spec))
        return self.client.get(self.get_spec_url(spec))

    def delete(self, spec, cdn=False):
        if cdn:
            return self.client.delete(self.get_spec_url_cdn(spec))
        return self.client.delete(self.get_url(spec))

    def test_ok(self):
        if not su_settings.DB_PRESET_ENABLED:
            return

        data = load_apistubs_yaml(su_settings.SPEC_FILES['ministubs'])
        self.post('vpnwn', data)
        response = self.get('vpnwn')

        self.assertEqual(data['paths'], json.loads(response.content)['paths'])
        response = self.delete('vpnwn')
        response = self.get('vpnwn')
        self.assertEqual(response.status_code, 404)




from django.views import View
from django.views.decorators.csrf import csrf_exempt
from django.http.response import HttpResponse

from apistubs.views.stub import BaseStubViewMixin
from apistubs.openapi.middleware import CheckOpenAPIMiddleware, OpenAPIValidationError

__all__ = (
    'StubForceView',
)


class StubForceView(BaseStubViewMixin, View):
    base_path = 'stubforce/'

    @csrf_exempt
    def dispatch(self, request, *args, **kwargs):
        middleware = CheckOpenAPIMiddleware(
            get_response=self.process,
            spec_from_path=True,
            explicit=True,
            base_path=self.base_path
        )
        try:
            return middleware(request, *args, **kwargs)
        except OpenAPIValidationError as e:
            return HttpResponse(e)



import yaml

from django.http import JsonResponse, HttpResponse
from django.views import View
from django.views.decorators.csrf import csrf_exempt

from apistubs.logging import RequestLog

__all__ = (
    'LogView',
)


class LogView(View):
    @csrf_exempt
    def dispatch(self, request, *args, **kwargs):
        return super(LogView, self).dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        env = kwargs.get('env', '')
        response_format = request.GET.get('format')
        if response_format == 'json':
            return JsonResponse({
                'log': RequestLog.get(env),
            })
        return HttpResponse(yaml.safe_dump(RequestLog.get(env)), content_type='text/plain')

    @csrf_exempt
    def delete(self, request, *args, **kwargs):
        env = kwargs.get('env', '')
        RequestLog.clear(env)
        return JsonResponse({})



from parse import Parser
from urllib.parse import parse_qs

from apistubs import settings as su_settings
from apistubs.helpers import get_path, replace_host, load_apistubs_yaml

__all__ = (
    'oas_find_path',
    'spec_point',
)


class ExtendedParser(Parser):
    def _handle_field(self, field):
        # handle as path parameter field
        field = field[1:-1]
        path_parameter_field = "{%s:PathParameter}" % field
        return super()._handle_field(path_parameter_field)


class PathParameter:
    name = "PathParameter"
    pattern = r"[^\/]+"

    def __call__(self, text):
        return text


parse_path_parameter = PathParameter()


def search(path_pattern, full_url_pattern) :
    extra_types = {parse_path_parameter.name: parse_path_parameter}
    p = ExtendedParser(path_pattern, extra_types)
    p._expression = '^' + p._expression + '$'
    return p.search(full_url_pattern)


def params_match(request, params):
    if not params or not request:
        return 1

    score = 0
    mack_params = [(key, value[0], ) for key, value in parse_qs(params).items()]
    for key, value in mack_params:
        for km, obj in (
            ('DATA.', request.POST, ),
            ('HEADER.', request.headers, ),
            ('', request.GET, ),
        ):
            if key.startswith(km):
                if obj.get(key[len(km):]) == value:
                    score += 1
                    break
                else:
                    return 0

    return score


def select_path(paths, path, request=None):
    max_path_size = 0
    result = None

    for path_pattern in paths:
        params = path_pattern.split('?')
        pattern = params[0]
        params = params[1] if len(params) > 1 else None

        params_match_score = params_match(request, params)
        if not params_match_score:
            continue

        path_size = len(pattern) + params_match_score
        if path_size > max_path_size and (
            pattern == path or
            search(pattern, path)
        ):
            result = path_pattern
            max_path_size = path_size

    return result


def oas_find_path(spec_name, path):
    spec_file = spec_point.get_spec_file(spec_name)
    spec = spec_point.get_data(spec_file)
    spec_paths = spec.get('paths', {}).keys()
    return select_path(spec_paths, path)


def response_from_spec(request, spec_name, pattern, requested_status, example_number):
    spec_file = spec_point.get_spec_file(spec_name)
    spec = spec_point.get_data(spec_file)
    responses = get_path(spec, 'paths', pattern, request.method.lower(), 'responses')
    if responses:
        for status in sorted(responses.keys()):
            if requested_status and int(status) != requested_status:
                continue

            examples = get_path(responses, status, 'content',  'application/json', 'examples')
            if not examples:
                example = get_path(responses, status, 'content',  'application/json', 'example')
                if example is not None:
                    examples = {'default': {'value': example}}

            if examples:
                headers = {}
                if status == '202':
                    location = get_path(responses, status, 'headers',  'Location', 'schema', 'example')
                    if location:
                        location = replace_host(location, request.get_host(), scheme=request.scheme)
                        headers['Location'] = location

                examples_keys = list(examples.keys())

                if isinstance(example_number, int):
                    if example_number + 1 > len(examples_keys):
                        example_number = 1
                    example = examples[examples_keys[example_number]]
                elif example_number is None:
                    example = list(examples.values())[0]
                else:
                    example = examples.get(example_number, {})

                content = example.get('value', {})
                return status, content, headers


class Spec():
    def __init__(self):
        self.data = {}

    def get_data(self, spec_file):
        if spec_file is None:
            return {}
        return load_apistubs_yaml(spec_file)

    def get_spec_file(self, spec_key):
        spec = su_settings.SPEC_FILES.get(spec_key)
        return spec


spec_point = Spec()



import json

from django.core.cache import cache

from apistubs import settings as su_settings
from apistubs.constants import METHODS
from apistubs.helpers import parse_preset_response, load_apistubs_yaml
from apistubs.spec import (
    oas_find_path,
    select_path,
    response_from_spec,
)

if su_settings.DB_PRESET_ENABLED:
    from apistubs.dbpreset.models import Mock

__all__ = (
    'StubResponse',
    'YamlSettings',
    'HeadersSettings',
    'DBSettings',
    'ComboSettings',
    'get_stub_response',
)


class Prompt:
    def __init__(self, value, env=False):
        self.value = value
        self.env = env

    def use_alias(self, status_aliases):
        selected_alias = None
        selected_alias_index = len(self.value)
        selected_alias_count = 0
        for status_alias in status_aliases:
            if isinstance(status_alias, str):
                alias = status_alias.split('-')[-1]
                if alias in self.value:
                    selected_alias_count += 1
                    index = self.value.index(alias)
                    if selected_alias_index > index:
                        selected_alias = status_alias
                        selected_alias_index = index

        if selected_alias_count > 1 and self.env is not None:
            self.value.pop(selected_alias_index)
            self.set_value(self.env, ' '.join(self.value))

        return selected_alias

    @classmethod
    def get_value(cls, env):
        return cache.get('PROMPT' + env)

    @classmethod
    def set_value(cls, env, value):
        return cache.set('PROMPT' + env, value, timeout=60 * 60 * 24 * 30)

    @classmethod
    def delete_value(cls, env):
        return cache.delete('PROMPT' + env)


class StubResponse:
    def __init__(
        self, status=200, content={}, headers=None, db_id=None, pattern=None, prompt=None
    ):
        self.status = status
        self.content = content
        if headers:
            self.headers = headers
        else:
            self.headers = {}
        self._db_id = db_id
        self.pattern = pattern
        self.prompt = prompt


class BaseSettingsSource:
    def __init__(self, spec_name, env='', path=None):
        self.spec_name = spec_name
        self.prompt = None
        self.env = env
        self.path = path
        self.values = self.load()

    def load(self):
        pass

    def set_prompt(self, value):
        if not value:
            return

        if isinstance(value, bytes):
            value = value.decode('utf-8')

        value = value.strip().replace('\n', ' ')
        value = [i.strip() for i in value.split(' ')]
        self.prompt = value

    @property
    def patterns(self):
        patterns = []

        for key in self.values:
            patterns.append(key.split('#')[-1])

        return patterns


class YamlSettings(BaseSettingsSource):
    def load(self):
        self.data_all = {}
        yaml_path = self.path
        if not yaml_path:
            return {}
        try:
            data = load_apistubs_yaml(yaml_path)
        except FileNotFoundError:
            return {}
        if not data:
            return {}
        self.set_prompt(data.get('PROMPT'))
        self.data_all = data
        return data.get(self.spec_name, {})


class HeadersSettings(BaseSettingsSource):
    def __init__(self, request):
        self.request = request
        self.values = self.load()

    @property
    def response(self):
        status = self.request.META.get('HTTP_STUB_RESPONSE_STATUS')
        content = self.request.META.get('HTTP_STUB_RESPONSE_CONTENT')
        headers = self.request.META.get('HTTP_STUB_RESPONSE_HEADERS')
        try:
            content = json.loads(content)
        except:
            return
        try:
            headers = json.loads(headers)
        except:
            return
        if status is None:
            return
        return StubResponse(status=int(status), content=content, headers=headers)


class DBSettings(BaseSettingsSource):
    def load(self):
        values = {}
        for response in Mock.objects.order_by('index').filter(spec_name=self.spec_name, env=self.env):
            values['#'.join([response.method, response.pattern])] = response.get_content()
        return values


class CookiesSettings(BaseSettingsSource):
    def __init__(self, request, env=''):
        self.request = request
        self.prompt = None
        self.env = env
        self.values = self.load()

    def load(self):
        paths = {}

        for cookie_name in self.request.COOKIES:
            for method in METHODS:
                if cookie_name.startswith('%s#' % method):
                    paths[cookie_name] = self.request.COOKIES[cookie_name]
                    paths.append(cookie_name.split('#')[-1])

        prompt = self.request.COOKIES.get('STUBS_PROMPT')
        if prompt:
            cache.set('PROMPT', prompt, 30)
        else:
            prompt = Prompt.get_value(self.env)
        self.set_prompt(prompt)
        return paths


class ComboSettings:
    def __init__(self, spec_name, request, env=''):
        self.request = request
        self.spec_name = spec_name
        self.use_db = su_settings.DB_PRESET_ENABLED
        self.prompt = None

        stubs_configs = su_settings.STUBS_CONFIG
        if not isinstance(stubs_configs, list):
            stubs_configs = [stubs_configs]
        self.yamls = [
            YamlSettings(spec_name, path=stubs_config)
            for stubs_config in stubs_configs
        ]
        for item in self.yamls:
            if item.prompt:
                self.prompt = Prompt(item.prompt)
                break

        self.headers = HeadersSettings(request)
        self.cookies = CookiesSettings(request, env=env)
        if self.cookies.prompt:
            self.prompt = Prompt(self.cookies.prompt, env=env)

        if self.use_db:
            self.db = DBSettings(spec_name, env=env)

    @property
    def patterns(self):
        pattens = []
        if self.use_db:
            pattens += self.db.patterns

        pattens += self.cookies.patterns
        for source in self.yamls:
            pattens += source.patterns

        return pattens

    def get_preset_response(self, pattern, path):
        sources = []
        if self.use_db:
            sources.append(self.db.values)
        sources.append(self.cookies.values)

        for source in self.yamls:
            sources.append(source.values)

        for source in sources:
            mp = '#'.join([self.request.method.lower(), path])
            if mp in source:
                return source[mp]
            mp = '#'.join([self.request.method.lower(), pattern])
            if mp in source:
                return source[mp]


def get_stub_response(spec_name, request, path, explicit=False, env=''):
    settings = ComboSettings(spec_name, request, env=env)

    response = settings.headers.response
    if response:
        return response

    pattern = oas_find_path(spec_name, path)
    if not pattern:
        pattern = select_path(settings.patterns, path, request=request)

    if not pattern:
        return

    preset_response = settings.get_preset_response(pattern, path)

    requested_status = None
    requested_example = None
    if preset_response:
        (
            requested_status,
            requested_example,
            requested_content,
            requested_headers,
        ) = parse_preset_response(preset_response, settings.prompt)
        if requested_status == 0:
            return
        if requested_status is not None and requested_content is not None:
            return StubResponse(
                status=requested_status,
                content=requested_content,
                headers=requested_headers,
                pattern=pattern,
                prompt=requested_example
            )

    if requested_status is None and explicit:
        return

    response = response_from_spec(
        request, spec_name, pattern,
        requested_status, requested_example
    )

    if response:
        status, content, headers = response
        return StubResponse(
            status=status,
            content=content,
            headers=headers,
            pattern=pattern,
            prompt=requested_example
        )

    if not requested_status:
        return

    return StubResponse(status=requested_status, pattern=pattern)



import json
import sys

from django.views import View
from django.http import HttpResponse, HttpResponseNotFound
from django.conf import settings as app_settings
from django.views.decorators.csrf import csrf_exempt

from apistubs import VERSION
from apistubs.stubs import get_stub_response
from apistubs.helpers import render_params
from apistubs.logging import RequestLog

__all__ = (
    'StubView',
    'IndexStubView',
)


class BaseStubViewMixin:
    base_path = 'stub/'

    def process(self, request, *args, **kwargs):
        spec_name = kwargs.get('spec', app_settings.PROJECT)
        env = kwargs.get('env', '')
        if self.base_path:
            path = request.path[request.path.find(self.base_path) + len(self.base_path) - 1:]
        else:
            path = request.path

        stub_response = get_stub_response(spec_name, request, path, env=env)
        if not stub_response:
            RequestLog.add_not_specified(
                service=spec_name, method=request.method, path=path,
                data=request.POST.dict(), params=request.GET.dict(), headers=dict(request.headers),
                env=env, request=request
            )
            return HttpResponseNotFound(json.dumps({'error': 'not_secified'}, indent=4, ensure_ascii=False))

        status, payload, headers = stub_response.status, stub_response.content, stub_response.headers

        RequestLog.add_success(
            service=spec_name, method=request.method, path=path,
            pattern=stub_response.pattern, status=status,
            content=payload, prompt=stub_response.prompt,
            data=request.POST.dict(), params=request.GET.dict(), headers=dict(request.headers),
            response_headers=headers, env=env, request=request
        )

        if not isinstance(payload, str):
            payload = json.dumps(payload, indent=4, ensure_ascii=False)

        response = HttpResponse(payload, status=status, content_type='application/json')
        for key in headers:
            header = str(headers[key]).strip()
            header = render_params(header, request)
            response[key] = header

        response['X-Stub-Mocked'] = 'on'
        response['X-Stub-Version'] = VERSION

        if spec_name:
            response['X-Stub-Service'] = spec_name
        return response


class StubView(BaseStubViewMixin, View):
    base_path = 'stub/'

    @csrf_exempt
    def dispatch(self, request, *args, **kwargs):
        return self.process(request, *args, **kwargs)


class IndexStubView(BaseStubViewMixin, View):
    base_path = None

    @csrf_exempt
    def dispatch(self, request, *args, **kwargs):
        return self.process(request, *args, **kwargs)



import os
from http import HTTPStatus
from typing import Callable
from urllib.parse import urlparse, urlsplit, urlunsplit

from django.conf import settings
from django.http.request import HttpRequest
from django.http.response import HttpResponse
from django.utils.http import urlencode
from openapi_core import OpenAPI, Spec
from openapi_core.contrib.django import \
    DjangoOpenAPIRequest as BaseDjangoOpenAPIRequest
from openapi_core.contrib.django.handlers import (
    DjangoOpenAPIErrorsHandler, DjangoOpenAPIValidRequestHandler)
from openapi_core.contrib.django.integrations import DjangoIntegration
from openapi_core.datatypes import RequestParameters
from openapi_core.deserializing.media_types.util import data_form_loads
from openapi_core.templating.paths.exceptions import (
    OperationNotFound,
    PathNotFound,
    ServerNotFound,
)
from openapi_core.templating.paths.finders import APICallPathFinder
from teamcity.messages import TeamcityServiceMessages
from werkzeug.datastructures import Headers, ImmutableMultiDict
from apistubs import settings as su_settings


CHECK_OPENAPI_SPEC = getattr(
    settings, 'CHECK_OPENAPI_SPEC',
    os.path.join(settings.BASE_DIR, 'docs', 'oas', 'api.json'),
)
CHECK_OPENAPI_PATHS = getattr(settings, 'CHECK_OPENAPI_PATHS', None)
EXCLUDE_OPENAPI_PATHS = getattr(settings, 'EXCLUDE_OPENAPI_PATHS', None)
SKIP_REQUEST = getattr(
    settings, 'CHECK_OPENAPI_SKIP_VALIDATE_REQUEST_STATUSES', [
        HTTPStatus.FORBIDDEN,
        HTTPStatus.BAD_REQUEST,
    ]
)
SKIP_RESPONSE = getattr(
    settings, 'CHECK_OPENAPI_SKIP_VALIDATE_RESPONSE_STATUSES', [
        HTTPStatus.FORBIDDEN,
        HTTPStatus.UNAUTHORIZED,
        HTTPStatus.NOT_FOUND,
        HTTPStatus.GONE,
        HTTPStatus.NOT_IMPLEMENTED,
        HTTPStatus.METHOD_NOT_ALLOWED,
    ]
)
BASE_HEADERS = {
    'User-Agent': 'Chrome/51.0.2704.103 Safari/537.36',
    'Referer': 'https://django.test/',
    'Authorization': 'Authorization',
    'X-CSRFToken': 'X-CSRFToken',
    'Accept-Language': 'en',
    'Cookie': '; '.join('='.join(i) for i in {
    }.items()),
}

__openapi_cache = {}

def get_finder(path):
    if path not in __openapi_cache:
        spec = Spec.from_file_path(path)
        finder = APICallPathFinder(
            spec, base_url=spec['servers'][0]['url']
        )
        openapi = OpenAPI(finder.spec)
        __openapi_cache[path] = finder, openapi
    return __openapi_cache[path]


messages = TeamcityServiceMessages()


class OpenAPIValidationError(Exception):
    pass


class DjangoOpenAPIRequest(BaseDjangoOpenAPIRequest):
    def __init__(self, request):
        self.request = request
        self.parameters = RequestParameters(
            path=self.request.openapi_path.path_result.variables,
            query=ImmutableMultiDict(self.request.GET),
            header=Headers(self.get_headers(request)),
            cookie=ImmutableMultiDict(dict(self.request.COOKIES)),
        )

        content = self.request.openapi_path.operation.get('requestBody', {}).get('content', {})
        self.request_pattern = urlparse(request.base_url).path + request.openapi_pattern
        self.request_mimetype = list(content.keys())[0] if content else None
        self.request_body = request._openapi_request_body
        self.base_url = request.base_url

    def get_headers(self, request):
        headers = {}
        headers.update(BASE_HEADERS)
        headers.update(dict(request.headers.items()))
        if headers.get('Content-Type', '').startswith('multipart/form-data'):
            headers['Content-Type'] = 'application/x-www-form-urlencoded'
        headers.pop('Content-Length', None)
        return headers

    @property
    def path_pattern(self) -> str:
        return self.request_pattern

    @property
    def host_url(self) -> str:
        return self.base_url

    @property
    def body(self):
        return self.request_body

    @property
    def content_type(self) -> str:
        if self.request_mimetype:
            return self.request_mimetype
        if self.request.content_type:
            return self.request.content_type.split(';')[0]
        return ''


class CheckOpenAPIMiddleware(DjangoIntegration):
    request_cls = DjangoOpenAPIRequest
    valid_request_handler_cls = DjangoOpenAPIValidRequestHandler
    errors_handler = DjangoOpenAPIErrorsHandler()

    def __init__(self, get_response: Callable[[HttpRequest], HttpResponse], spec_from_path=False, explicit=False, base_path=''):
        self.get_response = get_response
        self.spec_from_path = spec_from_path
        self.explicit = explicit
        self.base_path = base_path
        if spec_from_path:
            openapi_finder = None
            openapi = None
        else:
            openapi_finder, openapi = get_finder(CHECK_OPENAPI_SPEC)
        self.openapi_finder = openapi_finder
        super().__init__(openapi)

    def __call__(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:
        if self.spec_from_path:
            spec_name = kwargs['spec']
            spec = su_settings.SPEC_FILES.get(spec_name)
            if not spec:
                return HttpResponse('{}',  status=522)
            self.openapi_finder, self.openapi = get_finder(spec)

        try:
            path, *parts = urlsplit(request.path)[2:]
            path = path[path.index(self.base_path) + len(self.base_path):]
            url = urlunsplit(urlsplit(self.openapi_finder.base_url)[:2] + (path, ) + tuple(parts))
            openapi_path = self.openapi_finder.find(request.method.lower(), url)
        except (
            PathNotFound,
            OperationNotFound,
            ServerNotFound,  # TODO: investigate
        ):
            if self.explicit:
                return HttpResponse('{}',  status=522)
            return self.get_response(request, *args, **kwargs)
        else:
            request.base_url = self.openapi_finder.base_url
            request.openapi_path = openapi_path
            request.openapi_pattern = str(request.openapi_path.operation).split('#')[1]

            if CHECK_OPENAPI_PATHS is not None and request.openapi_pattern not in CHECK_OPENAPI_PATHS:
                return self.get_response(request, *args, **kwargs)

            if EXCLUDE_OPENAPI_PATHS is not None and request.openapi_pattern in EXCLUDE_OPENAPI_PATHS:
                return self.get_response(request, *args, **kwargs)

            content_type = request.META.get('CONTENT_TYPE')
            if content_type and content_type.startswith('multipart/form-data'):
                openapi_request_body = data_form_loads(request.body, boundary='BoUnDaRyStRiNg')
                openapi_request_body = urlencode(openapi_request_body).encode()
            else:
                openapi_request_body = request.body.decode('utf-8').encode()

            request._openapi_request_body = openapi_request_body

        response = self.get_response(request, *args, **kwargs)

        errors = []

        if response.status_code not in SKIP_REQUEST:
            request_unmarshal_result = self.unmarshal_request(request)
            if request_unmarshal_result.errors:
                errors.extend(request_unmarshal_result.errors)

        if (
            response.status_code not in SKIP_REQUEST and
            response.status_code not in SKIP_RESPONSE and
            response.status_code < 500
        ):
            response_unmarshal_result = self.unmarshal_response(request, response)
            if response_unmarshal_result.errors:
                errors.extend(response_unmarshal_result.errors)

        if errors:
            details = self.report(request, errors)
            if os.environ.get('TEAMCITY_VERSION'):
                self.report_teamcity(request, details)
            raise OpenAPIValidationError(details)

        return response

    def report(self,request, errors):
        errors = self.format_errors(errors)
        data = []
        data.append('[%s] %s' % (request.method, request.openapi_pattern,))

        for error in errors:
            data.append(error['title'])

        details = '\n'.join(data)
        return details

    def report_teamcity(self, request, details):
        messages.testStarted(request.openapi_pattern)
        messages.testFailed(request.openapi_pattern, '', details)
        messages.testFinished(request.openapi_pattern)

    def format_errors(self, errors):
        def format_openapi_error(error):
            if error.__cause__ is not None:
                error = error.__cause__
            return {
                'title': str(error),
                'type': str(type(error)),
            }
        data_errors = [format_openapi_error(err) for err in errors]
        return data_errors

"""
TBD
Remove after merge
https://github.com/python-openapi/openapi-core/pull/863
"""
from openapi_core.casting.schemas.casters import PrimitiveTypeCaster, BooleanCaster

class PatchBooleanCaster(PrimitiveTypeCaster[bool]):
    def validate(self, value) -> None:
        if isinstance(value, bool):
            return

        if value in [0, 1, '0', '1']:
            return

        if value.lower() not in ["false", "true"]:
            raise ValueError("not a boolean format")


BooleanCaster.validate = PatchBooleanCaster.validate



from __future__ import annotations

import json
from typing import Any, Dict, List, Tuple

from django import forms

from django.utils.html import escape
from django.core.cache import cache
from django.views import View
from django.views.decorators.csrf import csrf_exempt
from django.views.generic.edit import FormView
from django.http import (
    HttpResponse,
    HttpResponseRedirect,
)
from django.utils.safestring import SafeText

from apistubs import settings as su_settings
from apistubs.logging import RequestLog
from apistubs.stubs import YamlSettings, Prompt

if su_settings.DB_PRESET_ENABLED:
    from apistubs.dbpreset.models import Mock

__all__ = (
    'clean_prompt',
    'db_settings',
    'PromptView',
    'PromptAPIView',
)


def clean_prompt(prompt: bytes | str | None) -> List[str]:
    if not prompt:
        return []

    if isinstance(prompt, bytes):
        prompt = prompt.decode('utf-8')

    val = [p.strip() for p in prompt.replace(',', '').replace('\n', ' ').split(' ')]
    val = [p for p in val if p]
    return val


def db_settings(env: str) -> Dict[str, Dict[str, Any]]:
    settings: Dict[str, Dict[str, Any]] = {}
    if not su_settings.DB_PRESET_ENABLED:
        return settings

    for response in Mock.objects.order_by('index').filter(env=env):
        settings.setdefault(response.spec_name, {})
        settings[response.spec_name][
            '#'.join([response.method, response.pattern])
        ] = response.get_content()
    return settings


class PromptForm(forms.Form):
    default_key = 'default'
    special_fields = ['version']

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._initialize_properties(kwargs)
        self.base_fields: Dict[str, forms.Field] = {}
        self._create_prompt_field()
        self._create_choice_fields()
        super(PromptForm, self).__init__(*args, **kwargs)

    def _initialize_properties(self, kwargs: Dict[str, Any]) -> None:
        self.settings = kwargs.pop('settings')
        self.initial_prompt = kwargs.pop('prompt')
        self.prompt_query = kwargs.pop('prompt_query')
        self.default_data: Dict[str, Any] = {}
        self.anchors: List[str] = []

    def _create_prompt_field(self) -> None:
        self.base_fields[self.prompt_query] = forms.CharField(
            label='', widget=forms.Textarea, required=False, initial='\n'.join(self.initial_prompt)
        )

    def _create_choice_fields(self) -> None:
        prompt_index = 0
        for service, endpoints in self.settings.items():
            if service != service.lower() or service in self.special_fields:
                continue

            for index, path in enumerate(endpoints):
                if not isinstance(endpoints[path], dict):
                    continue

                prompts = self._get_prompts(endpoints[path])
                if len(prompts) < 2:
                    continue

                prompt_index += 1
                self._create_choice_field(service, path, index, prompt_index, prompts)

    def _get_prompts(self, endpoint_prompts: Dict[str, Any]) -> List[Tuple[str, str, str]]:
        prompts = []
        for prompt, payload in endpoint_prompts.items():
            parts = '{0}-{1}'.format(prompt, self.default_key).split('-', 2)
            prompts.append((
                parts[1],
                parts[0],
                json.dumps(payload, indent=2),
            ))
        return prompts

    def _create_choice_field(self, service: str, path: str, index: int, prompt_index: int, prompts: List[Tuple[str, str, str]]) -> None:
        field_name = f'{service}_{index}'
        initial = [p[0] for p in prompts if p[0] in self.initial_prompt]
        anchor = f'{service}-{path.replace("#", "")}'
        self.anchors.append(anchor)

        self.base_fields[field_name] = forms.ChoiceField(
            label=SafeText(
                f'<a href="#{anchor}" name="{anchor}">'
                f'{prompt_index}. [{service}] {path}</a>'
            ),
            label_suffix='',
            choices=[
                (p, SafeText(f'<span title=\"[{status}]\n{escape(payload)}\">{p}</span>'))
                for p, status, payload in prompts
            ],
            initial=initial[0] if initial else prompts[0][0],
            widget=forms.RadioSelect
        )
        self.default_data[field_name] = prompts[0][0]

    def clean_q(self) -> List[str]:
        return clean_prompt(self.cleaned_data[self.prompt_query])


class PromptView(FormView):
    template_name = 'apistubs/prompt.html'
    form_class = PromptForm
    cookie_name = 'STUBS_PROMPT'
    prompt_query = 'q'

    @csrf_exempt
    def dispatch(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        env = kwargs.get('env', '')
        self.db_settings = db_settings(env)
        return super(PromptView, self).dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        env = kwargs.get('env', '')
        response = super().get(request, *args, **kwargs)
        prompt = self.request.GET.get(self.prompt_query)
        if prompt is not None:
            response.set_cookie(self.cookie_name + env, prompt)
            if self.db_settings:
                cache.set('PROMPT' + env, prompt, timeout=60 * 60 * 24 * 30)
            else:
                cache.delete('PROMPT' + env)
        return response

    def form_valid(self, form):
        prompt = form.cleaned_data[self.prompt_query]
        if prompt == form.initial_prompt:
            prompt = [
                str(form.cleaned_data[key]) for key in form.cleaned_data
                if key != self.prompt_query and form.cleaned_data[key] != form.default_data[key]
            ]
        return HttpResponseRedirect('?%s=%s' % (self.prompt_query, '+'.join(prompt), ))

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        env = self.kwargs.get('env', '')
        kwargs['prompt'] = self._get_prompt(env)
        kwargs['prompt_query'] = self.prompt_query
        kwargs['settings'] = self._get_settings(env)
        return kwargs

    def _get_prompt(self, env):
        prompt = self.request.GET.get(self.prompt_query)
        if prompt is None:
            prompt = self.request.COOKIES.get(self.cookie_name + env, '')
            if self.db_settings:
                prompt = cache.get('PROMPT' + env, prompt)
        return clean_prompt(prompt)

    def _get_settings(self, env):
        if self.db_settings:
            return self.db_settings

        stubs_configs = su_settings.STUBS_CONFIG
        if not isinstance(stubs_configs, list):
            stubs_configs = [stubs_configs]
        stubs_configs = stubs_configs.copy()
        stubs_configs.reverse()
        settings = {}
        for stubs_config in stubs_configs:
            data_all = YamlSettings(None, path=stubs_config).data_all
            if not isinstance(data_all, dict):
                continue
            for service in data_all:
                if not isinstance(data_all[service], dict):
                    continue
                settings.setdefault(service, {})
                settings[service].update(data_all[service])
        return settings

    def get_context_data(self, **kwargs):
        env = self.kwargs.get('env', '')
        data = super(PromptView, self).get_context_data(**kwargs)
        if self.db_settings:
            data['yaml_path'] = 'from DB'
        else:
            data['yaml_path'] = 'from FILE'

        last_requests = []
        for item in RequestLog.get(env):
            if item['result'] == 'success':
                anchor = '%s-%s%s' % (item['service'], item['request']['method'], item.get('pattern', item['request']['path']),)
                name = '[%s] %s#%s' % (item['service'], item['request']['method'], item.get('pattern', item['request']['path']),)
                if anchor in data['form'].anchors:
                    last_request = (anchor, name, )
                    if last_request not in last_requests:
                        last_requests.append(last_request)

        data['env'] = env
        data['last_requests'] = last_requests
        return data


class PromptAPIView(View):
    @csrf_exempt
    def dispatch(self, request, *args, **kwargs):
        return super(PromptAPIView, self).dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        env = self.kwargs.get('env', '')
        return HttpResponse(Prompt.get_value(env) or '')

    def post(self, request, *args, **kwargs):
        env = self.kwargs.get('env', '')
        value = ' '.join(clean_prompt(request.body or ''))
        Prompt.set_value(env, value)
        return HttpResponse('')

    def delete(self, request, *args, **kwargs):
        env = self.kwargs.get('env', '')
        Prompt.delete_value(env)
        return HttpResponse('')



import json
import re
from urllib.parse import (
    urlsplit,
    urlunsplit,
)
from django.http import (
    HttpResponse,
    Http404,
)
from django.utils import timezone
from django import forms
from django.views import View
from django.views.generic import TemplateView
from django.views.decorators.csrf import csrf_exempt
from django.http import HttpResponse, Http404
from django.conf import settings as app_settings

from apistubs import settings as su_settings
from apistubs.helpers import get_path
from apistubs.spec import spec_point

__all__ = (
    'IndexView',
    'BrowserView',
    'OAuth2RedirectView',
    'SpecView',
)


if su_settings.DB_PRESET_ENABLED:
    from apistubs.dbpreset.models import Mock

class IndexView(TemplateView):
    template_name = 'apistubs/index.html'

    @csrf_exempt
    def dispatch(self, request, *args, **kwargs):
        return super(IndexView, self).dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        data = super(TemplateView, self).get_context_data(**kwargs)
        service = kwargs.get('spec', 'ministubs')
        data['service'] = service
        data['spec_url'] = None
        data['STATIC_URL'] = app_settings.STATIC_URL
        return data

    MOCK_MODEL_ENV_PREFIX = 'storage:'
    MOCK_MODEL_METHOD = 'spec'

    def get_env(self, name):
        return self.MOCK_MODEL_ENV_PREFIX + name

    def get_spec(self, name):
        value = Mock.objects.filter(env=self.get_env(name))
        if not value:
            return None, None
        item = value[0]
        return item.content, item

    def save_spec(self, name, value):
        if not value:
            return Mock.objects.filter(env=self.get_env(name)).delete()
        _, item = self.get_spec(name)
        if item:
            item.content = value
            return item.save()
        Mock.objects.create(
            index=-1,
            headers={},
            status=0,
            env=self.get_env(name),
            method=self.MOCK_MODEL_METHOD,
            content=value,
            created_at=timezone.now()
        )

    def post(self, request, *args, **kwargs):
        if not su_settings.DB_PRESET_ENABLED or 'spec' not in kwargs:
            raise Http404
        data = json.loads(request.body)
        self.save_spec(kwargs['spec'], request.body.decode())
        return HttpResponse()

    def delete(self, request, *args, **kwargs):
        if not su_settings.DB_PRESET_ENABLED or 'spec' not in kwargs:
            raise Http404
        self.save_spec(kwargs['spec'], None)
        return HttpResponse()


class BrowserView(TemplateView):
    template_name = 'apistubs/index.html'

    class URLForm(forms.Form):
        url = forms.URLField(required=False)

    def get_context_data(self, **kwargs):
        form = self.URLForm(data=self.request.GET)
        if not form.is_valid():
            raise Http404

        url = form.cleaned_data['url']

        data = super(TemplateView, self).get_context_data(**kwargs)
        data['browser_mode'] = True
        data['STATIC_URL'] = app_settings.STATIC_URL
        data['url'] = url
        return data


class OAuth2RedirectView(TemplateView):
    template_name = 'apistubs/oauth2-redirect.html'


class SpecView(View):
    def get(self, request, *args, **kwargs):
        spec_name = kwargs.get('spec', 'ministubs')
        data, item = self.get_spec(spec_name)
        if not data:
            if spec_name not in su_settings.SPEC_FILES:
                raise Http404
            spec_file = spec_point.get_spec_file(spec_name)
            data = spec_point.get_data(spec_file)

        self.process_data(data, spec_name, *args, **kwargs)
        data = json.dumps(data, indent=4, ensure_ascii=False)

        # TODO: fix images paths
        data = re.sub(re.compile('<figure>.*?</figure>'), '', data)

        response = HttpResponse(data, content_type='application/json')
        response['Access-Control-Allow-Origin'] = '*'
        return response

    MOCK_MODEL_ENV_PREFIX = 'storage:'

    def get_env(self, name):
        return self.MOCK_MODEL_ENV_PREFIX + name

    def get_spec(self, name):
        if not su_settings.DB_PRESET_ENABLED:
            return None, None
        value = Mock.objects.filter(env=self.get_env(name))
        if not value:
            return None, None
        item = value[0]
        data = item.content
        if isinstance(data, str):
            data = json.loads(data)
        return data, item

    def process_data(self, data, spec_name, *args, **kwargs):
        spec_name = kwargs.get('spec', 'ministubs')

        if (
            data['servers'] and
            isinstance(data['servers'], list) and
            isinstance(data['servers'][0], dict) and
            'url' in data['servers'][0] and
            isinstance(data['servers'][0]['url'], str)
        ):
            server_url_parsed = list(urlsplit(data['servers'][0]['url']))
            server_url_parsed[1] = self.request.get_host()
            data['servers'][0] = {
                'url': urlunsplit(server_url_parsed),
            }

        if su_settings.AUTHORIZATION_URL or su_settings.TOKEN_URL:
            oauth_flows = get_path(data, 'components', 'securitySchemes', 'oauth_2_0', 'flows')
            if oauth_flows:
                implicit = get_path(oauth_flows, 'implicit')
                if implicit and su_settings.AUTHORIZATION_URL:
                    implicit['authorizationUrl'] = su_settings.AUTHORIZATION_URL
                authorization_code = get_path(oauth_flows, 'authorizationCode')
                if authorization_code and su_settings.TOKEN_URL:
                    authorization_code['authorizationUrl'] = su_settings.AUTHORIZATION_URL
                    authorization_code['tokenUrl'] = su_settings.TOKEN_URL

        if spec_name == 'ministubs':
            specs = []
            if su_settings.DB_PRESET_ENABLED:
                items = Mock.objects.filter(env__startswith=self.MOCK_MODEL_ENV_PREFIX)
                specs += [item.env[len(self.MOCK_MODEL_ENV_PREFIX):] for item in items]
            specs += su_settings.SPEC_FILES.keys()
            specs.sort()
            data['paths']['/{service}/']['get']['parameters'][0]['description'] = '<b>Specifications:</b><ol>%s</ol>' % (
                ''.join([
                    '<li><a target="_blank" href="%s/">%s</a></li>' % (key, key,)
                    for key in specs
                ])
            )



import json
import yaml
import re

from django.conf import settings as app_settings
from django.http import (
    HttpResponse,
    JsonResponse,
    HttpResponseRedirect,
)
from django.views import View
from django.views.decorators.csrf import csrf_exempt

from apistubs.dbpreset.models import Mock
from apistubs.helpers import clear_comments

__all__ = (
    'SettingsView',
    'SpecSettingsView',
)


class SettingsView(View):
    @csrf_exempt
    def dispatch(self, request, *args, **kwargs):
        operation = request.POST.get('operation')
        if operation:
            env = request.POST.get('env', '')
            if not re.fullmatch(r'[-.\w]+', env):
                return HttpResponse('Invalid env')

            if operation == 'patch':
                data = request.POST.get('data')
                try:
                    preset = yaml.safe_load(data)
                except:
                    return HttpResponse('Invalid yaml')
                clear_comments(preset)

                self.operation_patch(preset, env)
                response = HttpResponseRedirect(request.path.replace('/settings/', f'/{env}/settings/'))
                response.set_cookie('stubs_env', env)
                return response
            elif operation == 'settings':
                return HttpResponseRedirect(request.path.replace('/settings/', f'/{env}/settings/'))
            elif operation == 'prompt':
                return HttpResponseRedirect(request.path.replace('/settings/', f'/{env}/prompt/'))

        return super(SettingsView, self).dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        env = kwargs.get('env', '')
        response_format = request.GET.get('format')
        responses = {}
        for response in Mock.objects.order_by('index').filter(env=env).all():
            responses.setdefault(response.spec_name, {})
            responses[response.spec_name][
                '#'.join([response.method, response.pattern])
            ] = response.get_content()

        if response_format == 'json':
            return HttpResponse(json.dumps(responses), content_type='text/json')

        return HttpResponse(yaml.safe_dump(responses, sort_keys=False), content_type='text/plain')

    def post(self, request, *args, **kwargs):
        env = kwargs.get('env', '')
        try:
            preset = json.loads(request.body)
        except json.JSONDecodeError:
            preset = yaml.safe_load(request.body)

        if not preset:
            preset = {}

        preset.pop('apistubs', None)
        clear_comments(preset)

        mocks = []
        index = 0
        for service in preset:
            if service.lower() != service:
                continue

            for pt in preset[service]:
                method, pattern = pt.split('#')
                preset_value = preset[service][pt]

                mocks.append(Mock(
                    index=index,
                    spec_name=service,
                    method=method,
                    pattern=pattern,
                    status=0,
                    content=Mock.prep_content(preset_value),
                    headers={},
                    env=env
                ))
                index += 1

        Mock.objects.filter(env=env).delete()
        Mock.objects.bulk_create(mocks)
        return HttpResponse()
        return JsonResponse(preset)

    def patch(self, request, *args, **kwargs):
        env = kwargs.get('env', '')
        try:
            preset = json.loads(request.body)
        except json.JSONDecodeError:
            preset = yaml.safe_load(request.body)

        clear_comments(preset)
        self.operation_patch(preset, env)
        return HttpResponse()
        return JsonResponse(preset)

    def operation_patch(self, preset, env):
        mocks = []
        index = 0
        for service in preset:
            for pt in preset[service]:
                method, pattern = pt.split('#')
                method = method.lower()
                preset_value = preset[service][pt]

                mocks.append(Mock(
                    index=index,
                    spec_name=service,
                    method=method,
                    pattern=pattern,
                    status=0,
                    content=preset_value,
                    headers={},
                    env=env
                ))
                index += 1

                Mock.objects.filter(
                    spec_name=service,
                    pattern=pattern,
                    method=method,
                    env=env
                ).delete()

        Mock.objects.bulk_create(mocks)
        return JsonResponse(preset)

    def delete(self, request, *args, **kwargs):
        env = kwargs.get('env', '')
        Mock.objects.filter(env=env).delete()
        return HttpResponse()


class SpecSettingsView(View):
    """
    Deprecated
    """
    @csrf_exempt
    def get(self, request, *args, **kwargs):
        spec_name = kwargs.get('spec', app_settings.PROJECT)

        responses = []
        for response in Mock.objects.order_by('id').filter(spec_name=spec_name).all():
            responses.append({
                'method': response.method,
                'pattern': response.pattern,
                'status': response.status,
                'headers': response.headers,
                'content': response.content,
            })
        return JsonResponse({
            'responses': responses,
        })

    @csrf_exempt
    def post(self, request, *args, **kwargs):
        # TODO: validation + documentation
        spec_name = kwargs.get('spec', app_settings.PROJECT)
        responses = json.loads(request.body)['responses']
        mocks = []
        for index, item in enumerate(responses):
            mocks.append(Mock(
                index=index,
                spec_name=spec_name,
                method=item['method'].lower(),
                pattern=item['pattern'],
                status=item['status'],
                content=item.get('content', {}),
                headers=item.get('headers', {}),
            ))

        Mock.objects.filter(spec_name=spec_name).delete()
        Mock.objects.bulk_create(mocks)
        return JsonResponse({
            'responses': responses,
        })

    @csrf_exempt
    def delete(self, request, *args, **kwargs):
        spec_name = kwargs.get('spec', app_settings.PROJECT)
        Mock.objects.filter(spec_name=spec_name).delete()
        return HttpResponse()



# Generated by Django 3.2.25 on 2024-06-30 16:41

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('dbpreset', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='mock',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='mock',
            name='env',
            field=models.CharField(db_index=True, default='', max_length=100),
        ),
    ]



import json
import sys

from django.conf import settings as app_settings
from django.utils.deprecation import MiddlewareMixin
from django.http import HttpResponse, JsonResponse

from apistubs import VERSION
from apistubs import settings as su_settings
from apistubs.stubs import get_stub_response
from apistubs.logging import RequestLog

__all__ = (
    'APIStubsMiddleware',
)


class APIStubsMiddleware(MiddlewareMixin):
    def process_request(self, request):
        if not su_settings.ENABLED:
            return

        if not su_settings.MIDDLEWARE_STUB_ENABLED:
            return

        if request.headers.get('X-Stubs-Mode') == 'skip':
            return

        env = ''
        marker = su_settings.MIDDLEWARE_STUB_COOKIE_MARKER
        if marker is True:
            marker = su_settings.MIDDLEWARE_STUB_COOKIE_MARKER_DEFAULT
        if marker:
            if marker not in request.COOKIES:
                return
            else:
                env = request.COOKIES[marker]
                env = env.strip()
                if env == 'root':
                    env = ''

        specs = [app_settings.PROJECT]
        if su_settings.MIDDLEWARE_SPECS:
            specs = su_settings.MIDDLEWARE_SPECS

        for spec in specs:
            stub_response = get_stub_response(spec, request, request.path, explicit=True, env=env)
            if not stub_response:
                continue

            status, payload, headers = stub_response.status, stub_response.content, stub_response.headers

            RequestLog.add_success(
                service=spec, method=request.method, path=request.path,
                pattern=stub_response.pattern, status=status,
                content=payload, prompt=stub_response.prompt,
                data=request.POST.dict(), params=request.GET.dict(), headers=dict(request.headers),
                response_headers=headers, request=request
            )

            if not isinstance(payload, str):
                payload = json.dumps(payload, indent=4, ensure_ascii=False)

            response = HttpResponse(payload, status=status, content_type='application/json')
            for header in headers:
                response[header] = headers[header]

            request.has_staff_ip = False
            response['X-Stub-Mocked'] = 'on'
            response['X-Stub-Version'] = VERSION

            return response



import os

from contextlib import contextmanager

from django.conf import settings as app_settings
from django.test.utils import override_settings

__all__ = (
    'settings',
    'default_app_config',
)


VERSION = '0.2.x'

default_settings = {
    'ENABLED': False,
    'MIDDLEWARE_STUB_ENABLED': False,
    'MIDDLEWARE_STUB_COOKIE_MARKER': None,
    'MIDDLEWARE_STUB_COOKIE_MARKER_DEFAULT': 'middleware_stubs_env',
    'SPEC_FILES': {},
    'AUTHORIZATION_URL': None,
    'TOKEN_URL': None,
    'EXTERNAL_DOCS': None,
    'MIDDLEWARE_SPECS': None,
    'STUBS_CONFIG': None,
    'STUB_FORCE_ENABLED': False,
    'DB_PRESET_ENABLED': False,
    'PRINT_INFO': True,
}


class Settings(object):
    app_name = __name__
    app_config = '%s.apps.APIStubsConfig' % __name__

    defaults = default_settings

    prefix = 'APISTUBS_'

    def __init__(self):
        self.reload()

    def reload(self):
        for k, v in self.get_settings().items():
            if hasattr(app_settings, self.prefix + k):
                value = getattr(app_settings, self.prefix + k, None)
            else:
                value = v
            setattr(self, k, value)

        self.SPEC_FILES.update({
            'ministubs': os.path.abspath(os.path.join(
                os.path.dirname(__file__), 'data', 'specs', 'ministubs.openapi.yaml'
            )),
        })

    @contextmanager
    def override(self, **kwargs):
        try:
            with override_settings(**kwargs):
                self.reload()
                yield
        finally:
            self.reload()

    def get_settings(self):
        data = self.defaults.copy()
        data.update({
            'STUBS_CONFIG': os.path.join(app_settings.BASE_DIR, '.stubs.yaml'),
            'DB_PRESET_ENABLED': self.get_setting('STUB_FORCE_ENABLED'),
        })
        return data

    def get_setting(self, key):
        return getattr(app_settings, self.prefix + key, self.defaults[key])

    def ready(self):
        pass



settings = Settings()
default_app_config = settings.app_config



import json
import yaml

from django.test import TestCase, override_settings
from django.urls import reverse

from apistubs import settings as su_settings
from apistubs import urls

__all__ = (
    'SettingsViewTests',
)


PROJECT = 'account'
STUB_SETTINGS = {
    'account': {
        'get#/requests': 200,
        'get#/requests': '200-ok',
        'get#/requests': {
            '300': {
                'status': 'ok',
            }
        },
    }
}


@override_settings(ROOT_URLCONF=urls, PROJECT=PROJECT)
class SettingsViewTests(TestCase):
    def get_url(self, env=False):
        if env:
            return reverse('settings_env', kwargs={'env': 'test'})
        return reverse('settings')

    def post(self, data, is_yaml=False, env=False):
        if is_yaml:
            data = yaml.dump(data)
            return self.client.patch(self.get_url(env=env), data=data, content_type='application/json')

        data = json.dumps(data)
        return self.client.post(self.get_url(env=env), data=data, content_type='application/json')

    def get(self, response_format='yaml', env=False):
        return self.client.get('%s?format=%s' %(self.get_url(env=env), response_format,))

    def delete(self):
        return self.client.delete(self.get_url())

    def test_ok(self):
        if not su_settings.DB_PRESET_ENABLED:
            return

        response = self.post(STUB_SETTINGS)
        response = self.post(STUB_SETTINGS, env=True)
        response = self.get()
        self.assertEqual(yaml.safe_load(response.content), STUB_SETTINGS)
        response = self.get(env=True)
        self.assertEqual(yaml.safe_load(response.content), STUB_SETTINGS)
        response = self.get(response_format='json')
        self.assertJSONEqual(response.content, STUB_SETTINGS)
        response = self.delete()
        response = self.get()
        self.assertJSONEqual(response.content, {})

    def test_ok_yaml(self):
        if not su_settings.DB_PRESET_ENABLED:
            return

        response = self.post(STUB_SETTINGS, is_yaml=True)
        response = self.post(STUB_SETTINGS, is_yaml=True, env=True)
        response = self.get()
        self.assertEqual(yaml.safe_load(response.content), STUB_SETTINGS)
        response = self.get( env=True)
        self.assertEqual(yaml.safe_load(response.content), STUB_SETTINGS)



import yaml
import json
from django.test import SimpleTestCase, TestCase, override_settings
from django.urls import reverse

from apistubs import urls
from apistubs.helpers import parse_preset_response
from apistubs.stubs import Prompt

__all__ = (
    'ParsePresetResponseTests',
    'PromptTests',
)


PROJECT = 'account'
APISTUBS = """
apistubs: 1.0.0

account:
  get#/service/accounts/{id}/accounts/:
    200-accounts_do_not_exist:
      accounts: []
    200-accounts_ok:
      accounts:
        - account_id: 500297762
          game: wows
          state: 3
          registered: True
    409-accounts_timeout:
      error: timeout
    500-accounts_error:
      error: server_error
"""


class ParsePresetResponseTests(SimpleTestCase):
    def test_ok(self):
        self.assertEqual(parse_preset_response('200'), (200, None, None, None,))
        self.assertEqual(parse_preset_response(200), (200, None, None, None,))
        self.assertEqual(parse_preset_response('200-key'), (200, 'key', None, None,))
        self.assertEqual(parse_preset_response(
            {200: {'status': 'ok'}}),
            (200, None, {'status': 'ok'}, None,)
        )
        self.assertEqual(parse_preset_response(
            {'200-full': {'status': 'ok'}}),
            (200, 'full', {'status': 'ok'}, None,)
        )
        self.assertEqual(parse_preset_response(
            {
                '200-full': {'status': 'ok'},
                '200-empty': {'status': 'ok'},
                '200-none': {'status': 'ok'},
            }, prompt=Prompt('empty')),
            (200, 'empty', {'status': 'ok'}, None,)
        )


@override_settings(ROOT_URLCONF=urls, PROJECT=PROJECT)
class PromptTests(TestCase):
    def test_ok(self):
        env = 'takeout'
        value = ['a2', 'a1', 'a3', 'b1', 'c1', 'c2', 'c4']
        prompt = Prompt(value, env=env)
        Prompt.set_value(env, value)

        alias = prompt.use_alias(['200-b0', '200-b1', '409-b2'])
        self.assertEqual(alias, '200-b1')

        alias = prompt.use_alias(['200-a0', '200-a1', '409-a2', '409-a3', '409-a4'])
        self.assertEqual(alias, '409-a2')
        self.assertEqual(Prompt.get_value(env), ' '.join(['a1', 'a3', 'b1', 'c1', 'c2', 'c4']))

        alias = prompt.use_alias(['200-a0', '200-a1', '409-a2', '409-a3', '409-a4'])
        self.assertEqual(alias, '200-a1')
        self.assertEqual(Prompt.get_value(env), ' '.join(['a3', 'b1', 'c1', 'c2', 'c4']))

        for x in range(4):
            alias = prompt.use_alias(['200-a0', '200-a1', '409-a2', '409-a3', '409-a4'])
            self.assertEqual(alias, '409-a3')
            self.assertEqual(Prompt.get_value(env), ' '.join(['a3', 'b1', 'c1', 'c2', 'c4']))

        value = ['b1', 'c1', 'c2', 'c4']
        prompt = Prompt(value, env=env)
        Prompt.set_value(env, ' '.join(value))
        alias = prompt.use_alias(['200-a0', '200-a1', '409-a2', '409-a3', '409-a4'])
        self.assertEqual(Prompt.get_value(env), ' '.join(['b1', 'c1', 'c2', 'c4']))
        self.assertEqual(alias, None)

        alias = prompt.use_alias(['200-c0', '409-c4', '200-c1', '409-c2', '409-c3'])
        self.assertEqual(alias, '200-c1')
        self.assertEqual(Prompt.get_value(env), ' '.join(['b1', 'c2', 'c4']))

        alias = prompt.use_alias(['200-c0', '409-c4', '200-c1', '409-c2', '409-c3'])
        self.assertEqual(alias, '409-c2')
        self.assertEqual(Prompt.get_value(env), ' '.join(['b1', 'c4']))

        for x in range(4):
            alias = prompt.use_alias(['200-c0', '409-c4', '200-c1', '409-c2', '409-c3'])
            self.assertEqual(alias, '409-c4')
            self.assertEqual(Prompt.get_value(env), ' '.join(['b1', 'c4']))

    def setup_stubs(self, env_url=False):
        data = yaml.safe_load(APISTUBS)
        if env_url:
            self.client.post(reverse('settings_env', kwargs={'env': 'test_env'}), data=data, content_type='application/json')
        else:
            self.client.post(reverse('settings'), data=data, content_type='application/json')

    def setup_prompt(self, prompt, env_url=False):
        if prompt is None:
            if env_url:
                self.client.delete(reverse('prompt_env', kwargs={'env': 'test_env'}))
            else:
                self.client.delete(reverse('prompt'))
        else:
            if env_url:
                self.client.post(reverse('prompt_env', kwargs={'env': 'test_env'}), data=prompt, content_type='text/plain')
            else:
                self.client.post(reverse('prompt'), data=prompt, content_type='text/plain')

    def get_prompt(self, env_url=False):
        if env_url:
            self.client.get(reverse('prompt_env', kwargs={'env': 'test_env'}))
        else:
            self.client.get(reverse('prompt'))

    def stub_request(self, env_url=False):
        url = reverse('stub', args=('account',))[:-1] + '/service/accounts/1/accounts/'
        if env_url:
            url = '/test_env' + url
        response = self.client.get(url)
        return json.loads(response.content)

    def _test_full_flow(self, env_url=False):
        self.setup_stubs(env_url=env_url)
        self.assertEqual(self.stub_request(env_url=env_url), {'accounts': []})

        self.setup_prompt('accounts_ok', env_url=env_url)
        self.assertEqual(self.stub_request(env_url=env_url), {
            'accounts': [{'account_id': 500297762,  'game': 'wows', 'registered': True,'state': 3}],
        })

        self.setup_prompt('accounts_error  accounts_timeout  accounts_ok', env_url=env_url)
        self.assertEqual(self.stub_request(env_url=env_url), {'error': 'server_error'})
        self.assertEqual(self.stub_request(env_url=env_url), {'error': 'timeout'})
        self.assertEqual(self.stub_request(env_url=env_url), {
            'accounts': [{'account_id': 500297762,  'game': 'wows', 'registered': True,'state': 3}],
        })

        self.setup_prompt(None, env_url=env_url)
        self.assertEqual(self.stub_request(env_url=env_url), {
            'accounts': [],
        })

    def test_full_flow(self):
        self._test_full_flow()

    def test_full_flow_env(self):
        self._test_full_flow(env_url=True)



import os
import yaml

from urllib.parse import (
    urlparse,
    urlunparse,
)
import json
from jinja2 import Environment, BaseLoader

__all__ = (
    'get_path',
    'replace_host',
    'render_params',
    'parse_preset_response',
    'clear_comments',
    'load_apistubs_yaml',
)


def get_path(data, *args):
    attr = data
    for arg in args:
        if attr and arg in attr:
            if not isinstance(attr, dict):
                return None
            attr = attr[arg]
        else:
            return None
    return attr


def replace_host(url, netloc, scheme=None):
    parsed_url = list(urlparse(url.strip()))
    parsed_url[1] = netloc
    if scheme:
        parsed_url[0] = scheme
    return urlunparse(parsed_url)


def render_params(value, request):
    if '{{' not in value or '}}' not in value:
        return value
    context = {}
    context.update(request.GET.dict())
    context.update(request.POST.dict())

    """
    fixed = {}
    for key in context:
        if '.' in key:
            fixed[key.replace('.', '_')] = context[key]
    import sys
    sys.stdout.write(json.dumps(fixed))
    context.update(fixed)
    """

    template = Environment(loader=BaseLoader).from_string(value)
    return template.render(**context)


def parse_preset_response(value, prompt=None):
    requested_status = None
    requested_example = None
    requested_content = None
    requested_headers = None

    if isinstance(value, dict):
        status_aliases = list(value.keys())
        status_alias = None
        if prompt:
            status_alias = prompt.use_alias(status_aliases)
        if status_alias is None:
            status_alias = status_aliases[0]
        payload = value[status_alias]
        if isinstance(status_alias, str):
            status_props = status_alias.split('-')
            requested_status = int(status_props[0])
            if len(status_props) > 1:
                requested_example = status_props[-1]
        else:
            requested_status = status_alias
        if isinstance(payload, dict):
            payload = payload.copy()
            requested_headers = payload.pop('HEADERS', None)
        requested_content = payload
    else:
        try:
            value = int(value)
        except ValueError:
            value = value.split('-')[:2]
            if len(value) == 2:
                status, example = value
                try:
                    example = int(example)
                except ValueError:
                    try:
                        payload = json.loads(example)
                        payload = payload.copy()
                    except:
                        pass
                    else:
                        if isinstance(payload, dict):
                            requested_headers = payload.pop('HEADERS', None)
                        requested_status = int(status)
                        requested_content = payload

                requested_status = int(status)
                requested_example = example
        else:
            requested_status = value

    return requested_status, requested_example, requested_content, requested_headers


def clear_comments(data, dep=0):
    # clear commented nodes ("_ ...")
    if dep == 3:
        return

    if not data or not isinstance(data, dict):
        return

    remove_nodes = []
    for app in data:
        if str(app)[0] == '_':
            remove_nodes.append(app)
        clear_comments(data[app], dep=dep + 1)

    for app in remove_nodes:
        data.pop(app)


__file_cache = {}
__file_cache_timestamp = {}


def load_apistubs_yaml(path):
    modefied = os.path.getctime(path)
    if __file_cache_timestamp.get(path) == modefied and path in __file_cache:
        return __file_cache[path]
    with open(path, 'r', encoding='utf-8') as f:
        if path.endswith('.json'):
            data = json.loads(f.read())
        else:
            data = yaml.safe_load(f)
    data.pop('apistubs', None)
    clear_comments(data)
    __file_cache[path] = data
    __file_cache_timestamp[path] = modefied
    return data



import os
import json
import yaml

from mock import ANY

from django.test import TestCase, override_settings
from django.urls import reverse

from apistubs import settings as su_settings
from apistubs.helpers import load_apistubs_yaml
from apistubs import urls

__all__ = (
    'StubViewTests',
    'StubForceViewTests',
    'SpecViewTests',
)

APP_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), '..' ))
PROJECT = 'account'


class ViewTestsMixin(object):
    def stub_url(self, service):
        return reverse('stub', args=(service,))

    def stub_env_url(self, env, service):
        return reverse('stub_env', args=(env, service,))

    def stub_default_url(self):
        return reverse('stub_default')

    def sample_url(self, service):
        return reverse('sample', args=(service,))

    def spec_url(self):
        return reverse('spec', args=(PROJECT,))

    def spec_default_url(self):
        return reverse('spec_default')

    def index_url(self):
        return reverse('index', args=(PROJECT,))

    def index_default_url(self):
        return reverse('index_default')

    def oauth2_redirect_url(self):
        return reverse('oauth2_redirect', args=('test',))

    def patch_spec(self):
        return su_settings.override(
            APISTUBS_SPEC_FILES={
                PROJECT: os.path.join(APP_ROOT, 'demo', 'tests.api.json'),
            },
            APISTUBS_STUBS_CONFIG=[
                os.path.join(APP_ROOT, 'demo', 'tests.stubs.json'),
                os.path.join(APP_ROOT, 'demo', 'tests.stubs.yaml'),
            ],
            APISTUBS_PRINT_INFO=False
        )

    def stub_request(
        self, path, expected_status, expected_content,
        service=None, specs=None, stubs_conf=None,
        default_url=False,
        env_url=False,
        method='get',
        post=None,
        headers={},
        expected_headers={},
        use_cookies_prompt=False,
        expected_log=None,
        content_type=None
    ):
        if service is None:
            service = PROJECT

        if specs is None:
            specs = {
                service: os.path.join(APP_ROOT, 'demo', 'tests.api.json'),
            }

        if stubs_conf is None:
            stubs_conf = [
                os.path.join(APP_ROOT, 'demo', 'tests.stubs.json'),
                os.path.join(APP_ROOT, 'demo', 'tests.stubs.yaml'),
            ]

        if default_url:
            base_path = self.stub_default_url()[:-1]
        elif env_url:
            base_path = self.stub_env_url('test_env', service)[:-1]
        else:
            base_path = self.stub_url(service)[:-1]

        path = base_path + path

        with su_settings.override(
            APISTUBS_SPEC_FILES=specs,
            APISTUBS_STUBS_CONFIG=stubs_conf,
            APISTUBS_PRINT_INFO=False
        ):
            if method == 'post':
                if content_type:
                    response = self.client.post(path, data=post, content_type=content_type, **headers)
                else:
                    response = self.client.post(path, data=post, **headers)
            else:
                response = self.client.get(path)

        self.assertEqual(response.status_code, expected_status)
        content = response.content.decode()
        if content:
            content = json.loads(content)
        self.assertEqual(content, expected_content)
        self.assertTrue(expected_headers.items() <= response.headers.items())

        if expected_log:
            self.client.delete(reverse('log'))
            self.client.delete(reverse('log_env', args=('test_env',)))

        if not su_settings.DB_PRESET_ENABLED:
            return

        with su_settings.override(
            APISTUBS_PRINT_INFO=False
        ):
            data = {}
            for stubs_path in stubs_conf:
                stubs_data = load_apistubs_yaml(stubs_path)
                data.update(stubs_data)

            prompt = data.get('PROMPT')
            if prompt:
                if use_cookies_prompt:
                    headers['Cookie'] = 'STUBS_PROMPT: %s' % prompt
                    from http.cookies import SimpleCookie
                    self.client.cookies = SimpleCookie({'STUBS_PROMPT': prompt})
                else:
                    if env_url:
                        self.client.post(reverse('prompt_env', kwargs={'env': 'test_env'}), data=prompt, content_type='text/plain')
                    else:
                        self.client.post(reverse('prompt'), data=prompt, content_type='text/plain')

            data = json.dumps(data)
            if env_url:
                self.client.post(reverse('settings_env', kwargs={'env': 'test_env'}), data=data, content_type='application/json')
            else:
                self.client.post(reverse('settings'), data=data, content_type='application/json')

            with su_settings.override(
                APISTUBS_SPEC_FILES=specs,
                APISTUBS_PRINT_INFO=False
            ):
                if method == 'post':
                    if content_type:
                        response = self.client.post(
                            path, data=post, content_type=content_type, **headers
                        )
                    else:
                        response = self.client.post(path, data=post, **headers)
                else:
                    response = self.client.get(path, **headers)
            self.assertEqual(response.status_code, expected_status)
            content = response.content.decode()
            if content:
                content = json.loads(content)
            self.assertEqual(content, expected_content)
            if expected_log:
                if env_url:
                    log_url = reverse('log_env', args=('test_env',))
                else:
                    log_url = reverse('log')

                response = self.client.get(log_url + '?format=json')
                content = response.content.decode()
                content = json.loads(content)
                self.assertEqual(content, expected_log)

                if not env_url:
                    log_url = reverse('log_env', args=('test_env',))
                else:
                    log_url = reverse('log')

                response = self.client.get(log_url + '?format=json')
                self.assertEqual(json.loads(response.content.decode()), {'log': []})


@override_settings(ROOT_URLCONF=urls, PROJECT=PROJECT)
class StubViewTests(ViewTestsMixin, TestCase):
    maxDiff = None

    def test_without_methods(self):
        self.stub_request('/', 404, {'error': 'not_secified'}, default_url=True)

    def test_does_not_exist_in_spec(self):
        self.stub_request('/does_not_exist_in_spec/', 404, {'error': 'not_secified'}, default_url=True)

    def test_does_not_exist_in_spec_but_in_conf(self):
        self.stub_request('/does_not_exist_in_spec_but_in_conf/', 200, {'status': 'one'},
                          default_url=True)

    def test_does_not_exist_in_spec_but_in_conf_use_cookies_prompt(self):
        self.stub_request(
            '/does_not_exist_in_spec_but_in_conf/', 200, {'status': 'one'},
            default_url=True, use_cookies_prompt=True,
            expected_log={'log': [{
                'result': 'success',
                'service': 'account',
                'prompt': 'ok',
                'request': {
                    'method': 'get',
                    'path': '/does_not_exist_in_spec_but_in_conf/',
                    'headers': ANY,
                    'params': {},
                    'data': {},
                    'body': None,
                },
                'response': {
                    'status': 200,
                    'content': {'status': 'one'},
                    'headers': {},
                },
            }]}
        )

    def test_default(self):
        self.stub_request('/family/api/children/list/', 202, {}, method='post')

    def test_default_env(self):
        self.stub_request('/family/api/children/list/', 202, {}, method='post', env_url=True)

    def test_example_number_predifined(self):
        self.stub_request('/auth/sessions/333/list/', 200, {'account_id': 12345, 'sessions': []})

    def test_env_stub(self):
        self.stub_request(
            '/auth/sessions/333/list/', 200, {'account_id': 12345, 'sessions': []}, env_url=True,
            expected_log={"log": [
                {
                    "result": "success",
                    "service": 'account',
                    "pattern": "/auth/sessions/{accountId}/list/",
                    "prompt": 2,
                    "request": {
                        "method": "get",
                        "path": "/auth/sessions/333/list/",
                        "data": {},
                        "headers": {
                            "Cookie": ""
                        },
                        "params": {},
                        'body': None,
                    },
                    "response": {
                        "status": "200",
                        "content": {
                            "account_id": 12345,
                            "sessions": []
                        },
                        "headers": {}
                    }
                }
            ]
        })

    def test_example_number_predifined(self):
        self.stub_request('/auth/sessions/333/list/', 200, {'account_id': 12345, 'sessions': []})

    def test_default_example(self):
        self.stub_request('/auth/sessions/333/list/', 200, {
            'account_id': 12345,
            'sessions': [ANY, ANY],
        }, service='spa2')

    def test_custom_status(self):
        self.stub_request('/auth/sessions/333/list/', 500, {}, service='spa3')

    def test_custom_status_and_response(self):
        self.stub_request('/auth/sessions/333/list/', 500, '', service='spa4')

    def test_template_response(self):
        query = (
            'client_id=929991016477-0t10462r6ifugrcqoeieeqokfqklva3o.apps.googleusercontent.com&'
            'redirect_uri=https%3A%2F%2Fexample.net%2Fpersonal%2Fcredentials%2Fexternal%2Fbind%2Fprocess%2F&'
            'response_type=code&'
            'scope=profile+email+https%3A%2F%2Fwww.googleapis.com%2Fauth%2Fuser.birthday.read&'
            'state=google%3A9565d1665544b2ed0a8d5a4a9a599191'
        )
        self.stub_request('/auth/sessions/333/list/?' + query, 301, {}, service='spa5', expected_headers={
            'Location': (
                'https://example.net/personal/credentials/external/bind/process/&state=google:9565d1665544b2ed0a8d5a4a9a599191&'
                'scope=profile%20email%20https%3A//www.googleapis.com/auth/user.birthday.read&'
                'authuser=1&'
                'prompt=none&'
                'code=111111'
            ),
        })

    def test_custom_third(self):
        self.stub_request('/custom/', 409, {}, service='third')

    def test_skipped(self):
        self.stub_request('/skipped/', 404, {'error': 'not_secified'}, service='third')

    def test_skipped_with_prompt(self):
        self.stub_request('/skipped2/', 404, {'error': 'not_secified'}, service='third')

    def test_skipped_with_prompt_env(self):
        self.stub_request('/skipped2/', 404, {'error': 'not_secified'}, service='third', env_url=True)

    def test_skipped_with_cookie_prompt(self):
        self.stub_request(
            '/skipped2/', 404, {'error': 'not_secified'}, service='third',
            use_cookies_prompt=True, method='post',
            post=json.dumps({'key_post': 'value_post'}),
            content_type='application/json',
            expected_log={'log': [{
                'result': 'not_specified',
                'service': 'third',
                'request': {
                    'method': 'post',
                    'path': '/skipped2/',
                    'headers': ANY,
                    'params': {},
                    'data': {},
                    'body': {'key_post': 'value_post'},
                },
            }]}
        )

    def test_custom_response(self):
        self.stub_request(
            '/realm/detect/', 409, {'realm': 'one'},
            expected_headers={'Etag': '32423412342'}
        )

    def test_request_view(self):
        self.stub_request('/personal/account/birthday/', 202, {}, method='post')

    def test_parametrize(self):
        self.stub_request('/parametrize/?key=value&key2=value2', 200, {'status': 'ok'}, method='get')

    def test_parametrize_data(self):
        self.stub_request(
            '/parametrize/?key=value', 200, {'status': 'ok'}, method='post',
            post={'key_post': 'value_post'},
            headers={'HTTP_X_TRACKING_ID': '500zxc'}
        )


@override_settings(ROOT_URLCONF=urls, PROJECT=PROJECT)
class StubForceViewTests(ViewTestsMixin, TestCase):
    def stub_url(self, service):
        return reverse('stub_force', args=(service,))

    def test_default(self):
        self.stub_request('/family/api/children/list/', 202, {}, method='post')

    def test_default_env(self):
        self.stub_request('/family/api/children/list/', 202, {}, method='post', env_url=True)

    def test_example_number_predifined(self):
        self.stub_request('/auth/sessions/333/list/', 200, {'account_id': 12345, 'sessions': []})

    def test_env_stub(self):
        self.stub_request(
            '/auth/sessions/333/list/', 200, {'account_id': 12345, 'sessions': []}, env_url=True,
            expected_log={"log": [
                {
                    "result": "success",
                    "service": 'account',
                    "pattern": "/auth/sessions/{accountId}/list/",
                    "prompt": 2,
                    "request": {
                        "method": "get",
                        "path": "/auth/sessions/333/list/",
                        "data": {},
                        "headers": {
                            "Cookie": ""
                        },
                        "params": {},
                        'body': None,
                    },
                    "response": {
                        "status": "200",
                        "content": {
                            "account_id": 12345,
                            "sessions": []
                        },
                        "headers": {}
                    }
                }
            ]
        })

    def test_example_number_predifined(self):
        self.stub_request('/auth/sessions/333/list/', 200, {'account_id': 12345, 'sessions': []})

    def test_default_example(self):
        self.stub_request('/auth/sessions/333/list/', 200, {
            'account_id': 12345,
            'sessions': [ANY, ANY],
        }, service='spa2')

    def test_custom_status(self):
        self.stub_request('/auth/sessions/333/list/', 500, {}, service='spa3')

    def test_custom_status_and_response(self):
        self.stub_request('/auth/sessions/333/list/', 500, '', service='spa4')

    def test_template_response(self):
        query = (
            'client_id=929991016477-0t10462r6ifugrcqoeieeqokfqklva3o.apps.googleusercontent.com&'
            'redirect_uri=https%3A%2F%2Fexample.net%2Fpersonal%2Fcredentials%2Fexternal%2Fbind%2Fprocess%2F&'
            'response_type=code&'
            'scope=profile+email+https%3A%2F%2Fwww.googleapis.com%2Fauth%2Fuser.birthday.read&'
            'state=google%3A9565d1665544b2ed0a8d5a4a9a599191'
        )
        self.stub_request('/auth/sessions/333/list/?' + query, 301, {}, service='spa5', expected_headers={
            'Location': (
                'https://example.net/personal/credentials/external/bind/process/&state=google:9565d1665544b2ed0a8d5a4a9a599191&'
                'scope=profile%20email%20https%3A//www.googleapis.com/auth/user.birthday.read&'
                'authuser=1&'
                'prompt=none&'
                'code=111111'
            ),
        })

    def test_skipped_with_prompt(self):
        self.stub_request('/skipped2/', 404, {'error': 'not_secified'}, service='third')

    def test_skipped_with_prompt_env(self):
        self.stub_request('/skipped2/', 404, {'error': 'not_secified'}, service='third', env_url=True)

    def test_custom_response(self):
        self.stub_request(
            '/realm/detect/', 409, {'realm': 'one'},
            expected_headers={'Etag': '32423412342'}
        )

    def test_request_view(self):
        self.stub_request('/personal/account/birthday/?birthday=birthday', 202, {}, method='post')


@override_settings(ROOT_URLCONF=urls, PROJECT=PROJECT)
class SpecViewTests(ViewTestsMixin, TestCase):
    def test_index(self):
        response = self.client.get(self.oauth2_redirect_url())
        self.assertEqual(response.status_code, 200)

        with self.patch_spec():
            self.client.get(self.index_url())
            self.client.get(self.index_default_url())

    def test_spec(self):
        with self.patch_spec():
            for url in [
                self.spec_url(),
                self.spec_default_url()
            ]:
                response = self.client.get(url)

                self.assertEqual(response.status_code, 200)
                content = json.loads(response.content)
                self.assertEqual(content, {
                    'openapi': ANY,
                    'info': ANY,
                    'servers': ANY,
                    'paths': ANY,
                    'tags': ANY,
                    'components': ANY,
                })



import sys
import json

from django.core.cache import cache

from apistubs import settings as su_settings

__all__ = (
    'RequestLog',
)


BRIGHT_YELLOW = '\033[93m'
BRIGHT_RED = '\033[91m'
RESET = '\033[0m'


def _get_request_body(request):
    try:
        # TODO: set-up proper Exceptions
        body = json.loads(request.body)
    except:
        return None
    return body


class RequestLog:
    MAX = 20
    CACHE_KEY = 'LOG'

    @classmethod
    def get(cls, env):
        value = cache.get(cls.CACHE_KEY + env)
        if not value:
            value = []
        return value

    @classmethod
    def add_success(
        cls, service='default', method='get', path='/',
        pattern=None, status=200, content={}, prompt=None, data={}, headers={},
        response_headers={}, params={}, env='', request=None
    ):
        msg = {
            'result': 'success',
            'service': service,
        }
        if pattern and path != pattern:
            msg['pattern'] = pattern
        if prompt:
            msg['prompt'] = prompt
        msg['request'] = {
            'method': method.lower(),
            'path': path,
            'data': data,
            'headers': headers,
            'params': params,
            'body': _get_request_body(request),
        }
        msg['response'] = {
            'status': status,
            'content': content,
            'headers': response_headers,
        }
        cls.add(msg, env)

        if su_settings.PRINT_INFO:
            sys.stdout.write(
                BRIGHT_YELLOW +
                '[STUB] {}: {}#{}\n'.format(service, method.lower(), pattern) +
                RESET
            )

    @classmethod
    def add_not_specified(
        cls, service='default', method='get', path='/', data={},
        headers={}, params={}, env='', request=None
    ):
        msg = {
            'result': 'not_specified',
            'service': service,
            'request': {
                'method': method.lower(),
                'path': path,
                'data': data,
                'headers': headers,
                'params': params,
                'body': _get_request_body(request),
            }
        }
        cls.add(msg, env)

        if su_settings.PRINT_INFO:
            sys.stdout.write((
                BRIGHT_RED +
                '[STUB][NOT_FOUND] {s}: {m}#{p}' +
                BRIGHT_YELLOW +
                '\n[StubSampler to fill-up Stubs.YAML] ' +
                '$ ministubs sample --snapshots [ANY] --service {s} --path {p}' +
                RESET + '\n'
            ).format(s=service, m=method.lower(), p=path))

    @classmethod
    def add(cls, item, env):
        value = cls.get(env)
        value.insert(0, item)
        value = value[:cls.MAX]
        cache.set(cls.CACHE_KEY + env, value, timeout=60 * 60 * 24 * 30)

    @classmethod
    def clear(cls, env):
        cache.delete(cls.CACHE_KEY + env)



from django.db import models


class Mock(models.Model):
    index = models.IntegerField()
    headers = models.JSONField()
    content = models.JSONField()
    status = models.IntegerField()
    method = models.CharField(max_length=6)
    pattern = models.CharField(max_length=1000)
    spec_name = models.CharField(db_index=True, max_length=100)
    env = models.CharField(db_index=True, max_length=100, default='')
    created_at = models.DateTimeField(auto_now_add=True)

    def get_content(self):
        if isinstance(self.content, list):
            return dict(self.content)
        return self.content

    @staticmethod
    def prep_content(content):
        if isinstance(content, dict):
            return [list(i) for i in content.items()]
        return content



from django.urls import re_path as url

from apistubs import settings as su_settings

app_name = 'apistubs'
urlpatterns = []

if su_settings.ENABLED:
    from apistubs.views.common import (
        IndexView,
        BrowserView,
        OAuth2RedirectView,
        SpecView,
    )
    from apistubs.views.stub import (
        StubView,
    )
    from apistubs.views.prompt import (
        PromptView,
        PromptAPIView,
    )
    from apistubs.views.logging import (
        LogView,
    )

    urlpatterns += [
        url(r'^browser/$', BrowserView.as_view(), name='browser', kwargs={'SSL': 3}),
    ]

    urlpatterns += [
        url(r'^$', IndexView.as_view(), name='index_default'),
    ]

    urlpatterns += [
        url(r'^stub/', StubView.as_view(), name='stub_default'),
        url(r'^api.json$', SpecView.as_view(), name='spec_default'),
        url(r'^prompt/$', PromptView.as_view(), name='prompt_index'),
        url(r'^prompt/api/$', PromptAPIView.as_view(), name='prompt'),
        url(r'^log/$', LogView.as_view(), name='log'),
        url(r'^(?P<env>[-.\w]+)/prompt/$', PromptView.as_view(), name='prompt_env_index'),
        url(r'^(?P<env>[-.\w]+)/prompt/api/$', PromptAPIView.as_view(), name='prompt_env'),
        url(r'^(?P<env>[-.\w]+)/log/$', LogView.as_view(), name='log_env'),
    ]

    if su_settings.DB_PRESET_ENABLED:
        from apistubs.views.settings import SettingsView
        urlpatterns += [
            url(r'^settings/', SettingsView.as_view(), name='settings'),
            url(r'^(?P<env>[-.\w]+)/settings/', SettingsView.as_view(), name='settings_env'),
        ]

    urlpatterns += [
        url(r'^(?P<spec>[-.\w]+)/$', IndexView.as_view(), name='index'),
        url(r'^(?P<spec>[-.\w]+)/api.json$', SpecView.as_view(), name='spec'),
        url(r'^(?P<spec>[-.\w]+)/oauth2-redirect.html$', OAuth2RedirectView.as_view(), name='oauth2_redirect'),
    ]

    urlpatterns += [
        url(r'^(?P<spec>[-.\w]+)/stub/', StubView.as_view(), name='stub'),
        url(r'^(?P<env>[-.\w]+)/(?P<spec>[-.\w]+)/stub/', StubView.as_view(), name='stub_env'),
    ]

    if su_settings.STUB_FORCE_ENABLED:
        from apistubs.openapi.stubforce import StubForceView
        urlpatterns += [
            url(r'^(?P<spec>[-.\w]+)/stubforce/', StubForceView.as_view(), name='stub_force'),
            url(r'^(?P<env>[-.\w]+)/(?P<spec>[-.\w]+)/stubforce/', StubForceView.as_view(), name='stub_force_env'),
        ]



import json
import yaml

from django.test import TestCase, override_settings
from django.urls import reverse

from apistubs import settings as su_settings
from apistubs import urls

__all__ = (
    'PromptViewTests',
)


PROJECT = 'account'

TEST_PROMPT = 'a1wefg2345234 asdfswdfg345'
TEST_PROMPT_ENV = 'asfgsd sfgsdg sdfhg'

@override_settings(ROOT_URLCONF=urls, PROJECT=PROJECT)
class PromptViewTests(TestCase):
    def get_url(self, env=False, prompt=None):
        if env:
            url = reverse('prompt_env', kwargs={'env': 'test'})
        else:
            url = reverse('prompt')
        return url

    def get_index_url(self, env=False, prompt=None):
        if env:
            url = reverse('prompt_env_index', kwargs={'env': 'test'})
        else:
            url = reverse('prompt_index')
        if prompt:
            url = url + '?q=' + prompt
        return url

    def test_api_ok(self):
        response = self.client.get(self.get_url(env=False))
        self.assertEqual(response.content.decode(), '')

        response = self.client.post(self.get_url(env=False), TEST_PROMPT, content_type='text/plain')
        response = self.client.post(self.get_url(env=True), TEST_PROMPT_ENV, content_type='text/plain')

        response = self.client.get(self.get_url(env=False))
        self.assertEqual(response.content.decode(), TEST_PROMPT)

        response = self.client.get(self.get_url(env=True))
        self.assertEqual(response.content.decode(), TEST_PROMPT_ENV)

        response = self.client.delete(self.get_url(env=False))
        response = self.client.delete(self.get_url(env=True))

        response = self.client.get(self.get_url(env=False))
        self.assertEqual(response.content.decode(), '')

        response = self.client.get(self.get_url(env=True))
        self.assertEqual(response.content.decode(), '')

    def test_ok(self):
        response = self.client.get(self.get_index_url(env=False, prompt=TEST_PROMPT))
        self.assertContains(response, TEST_PROMPT.replace(' ', '\n'))

        response = self.client.get(self.get_index_url(env=True, prompt=TEST_PROMPT_ENV))
        self.assertContains(response, TEST_PROMPT_ENV.replace(' ', '\n'))

        response = self.client.get(self.get_index_url(env=True))
        self.assertContains(response, TEST_PROMPT_ENV.replace(' ', '\n'))

        response = self.client.get(self.get_index_url(env=False))
        self.assertContains(response, TEST_PROMPT.replace(' ', '\n'))
